import argparse

import numpy as np
import pyperf
from utils import (
    geometric_matrix,
    machol_wien_matrix,
    random_machol_wien_matrix,
    uniform_matrix,
)


def get_solvers():
    from lapjv import lapjv as lapjv_lap

    # from lapjv_noinit import lapjv as lapjv_noinit_lap
    from lapsolver import solve_dense
    from scipy.optimize import linear_sum_assignment as scipy_lap
    from laptools.lap import solve as laptools_lap

    return {
        "scipy": scipy_lap,
        "lapjv": lapjv_lap,
        # "lapjv_noinit": lapjv_noinit_lap,
        "lapsolver": solve_dense,
        "laptools": laptools_lap,
    }


def time_func(n_inner_loops, solver, shape, type):
    # Note: If no matrix type is indicated, then the matrix is uniformly random
    if type == "uniform":
        cost_matrix = uniform_matrix(shape)
    if type == "geometric":
        cost_matrix = geometric_matrix(shape)
    elif type == "MW":
        cost_matrix = machol_wien_matrix(shape)
    elif type == "random_MW":
        cost_matrix = random_machol_wien_matrix(shape)
    else:
        cost_matrix = np.random.random(shape)

    t0 = pyperf.perf_counter()
    for i in range(n_inner_loops):
        solver(cost_matrix)
    return pyperf.perf_counter() - t0


def get_bench_name(size, type, solver_name):
    return "{}-{}-{}".format(size, type, solver_name)


def parse_args(benchopts):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--min-size-pow",
        type=int,
        metavar="POW",
        default=1,
        help="Smallest test matrix will be size 2^POW by 2^POW.",
    )
    parser.add_argument(
        "--max-size-pow",
        type=int,
        metavar="POW",
        default=2,
        help="Largest test matrix will be size 2^POW by 2^POW.",
    )
    parser.add_argument(
        "--matrix-type",
        type=str,
        metavar="X",
        default="random",
        help="The matrix is of type X.",
    )
    return parser.parse_args(benchopts)


def add_cmdline_args(cmd, args):
    cmd.append("--")
    cmd.extend(args.benchopts)


def main():
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.argparser.add_argument("benchopts", nargs="*")
    args = parse_args(runner.parse_args().benchopts)

    solvers = get_solvers()
    sizes = 2 ** np.arange(args.min_size_pow, args.max_size_pow + 1)
    type = args.matrix_type
    for size in sizes:
        for solver_name, solver_func in solvers.items():
            bench_name = get_bench_name(size, type, solver_name)
            runner.bench_time_func(
                bench_name, time_func, solver_func, (size, size), type
            )


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from laptools import clap

# fmt: off
cost_matrix_0 = [[4, 1, 3],
                 [2, 0, 5],
                 [3, 2, 2]]

cost_matrix_1 = [[4, 1, 3, 6],
                 [2, 0, 5, 7],
                 [3, 2, 2, 8]]

cost_matrix_2 = [[4, 1, 3],
                 [2, 0, 5],
                 [3, 2, 2],
                 [6, 7, 8]]

cost_matrix_3 = [[np.inf, np.inf, np.inf, np.inf,      3],
                 [     7, np.inf,     23, np.inf, np.inf],
                 [    17,     24, np.inf, np.inf, np.inf],
                 [np.inf,      6,     13,     20, np.inf]]

cost_matrix_4 = [[4, 1, 3, np.inf],
                 [2, 0, 5, np.inf],
                 [3, 2, 2, np.inf]]

cost_matrix_5 = [[4, 1, 3],
                 [2, 0, 5],
                 [np.inf, np.inf, np.inf],
                 [3, 2, 2],
                 [6, 7, 8]]

global_cost_0 = np.array([[6, 5, 6],
                          [5, 6, 9],
                          [6, 7, 5]])

global_cost_1 = np.array([[6, 5, 6, 8],
                          [5, 6, 9, 10],
                          [6, 7, 5, 11]])

global_cost_2 = np.array([[6, 5, 6],
                          [5, 6, 9],
                          [6, 7, 5],
                          [8, 11, 11]])

global_cost_3 = np.array([[np.inf, np.inf, np.inf, np.inf,     47],
                          [    47, np.inf,     49, np.inf, np.inf],
                          [    49,     47, np.inf, np.inf, np.inf],
                          [np.inf,     49,     47,     54, np.inf]])

global_cost_4 = np.array([[6, 5, 6, np.inf],
                          [5, 6, 9, np.inf],
                          [6, 7, 5, np.inf]])

global_cost_5 = np.array([[6, 5, 6],
                          [5, 6, 9],
                          [np.inf, np.inf, np.inf],
                          [6, 7, 5],
                          [8, 11, 11]])

# New tests:
cost_matrix_6 = [[3, 8, 1, 3, 8],
                 [2, 0, 6, 8, 9],
                 [6, 7, 5, 8, 5],
                 [6, 3, 8, 2, 0],
                 [2, 9, 6, 6, 3]]

global_cost_6 = np.array([[13, 20, 10, 10, 17],
                          [15, 10, 18, 18, 21],
                          [12, 15, 10, 11, 10],
                          [17, 16, 18, 10, 10],
                          [10, 19, 15, 13, 12]])

cost_matrix_7 = [[4, 1, 3],
                 [2, np.inf, np.inf],
                 [3, np.inf, np.inf]]

global_cost_7 = np.array([[np.inf, np.inf, np.inf],
                          [np.inf, np.inf, np.inf],
                          [np.inf, np.inf, np.inf]])

# fmt: on


class TestClap:
    # @pytest.mark.parametrize(
    #     "i, j, cost_matrix, expected",
    #     [
    #         (0, 0, cost_matrix_0, 6),
    #         (0, 1, cost_matrix_0, 5),
    #         (0, 2, cost_matrix_0, 6),
    #         (1, 0, cost_matrix_0, 5),
    #         (1, 1, cost_matrix_0, 6),
    #         (1, 2, cost_matrix_0, 9),
    #         (2, 0, cost_matrix_0, 6),
    #         (2, 1, cost_matrix_0, 7),
    #         (2, 2, cost_matrix_0, 5),
    #         (0, 3, cost_matrix_1, 8),
    #         (1, 3, cost_matrix_1, 10),
    #         (2, 3, cost_matrix_1, 11),
    #         (3, 0, cost_matrix_2, 8),
    #         (3, 1, cost_matrix_2, 11),
    #         (3, 2, cost_matrix_2, 11),
    #     ],
    # )
    # def test_clap_cost(self, i, j, cost_matrix, expected):
    #     """Verify clap.cost works on small examples computable by hand."""
    #     assert clap.cost(i, j, cost_matrix) == expected

    @pytest.mark.parametrize(
        "cost_matrix, expected_global_costs",
        [
            (cost_matrix_0, global_cost_0),
            (cost_matrix_1, global_cost_1),
            (cost_matrix_2, global_cost_2),
            (cost_matrix_3, global_cost_3),
            (cost_matrix_4, global_cost_4),
            (cost_matrix_5, global_cost_5),
            (cost_matrix_6, global_cost_6),
            (cost_matrix_7, global_cost_7),
        ],
    )
    def test_clap_costs(self, cost_matrix, expected_global_costs):
        """Verify clap.constrained_lsap_costs works on small examples."""
        assert clap.costs(cost_matrix).tolist() == expected_global_costs.tolist()

        # num_rows, num_cols = np.array(cost_matrix).shape
        # for i in range(num_rows):
        #     for j in range(num_cols):
        #         assert clap.cost(i, j, cost_matrix) == expected_global_costs[i, j]



import numpy as np

from _augment import _solve

from . import lap
from ._util import one_hot


def costs(cost_matrix):
    """Solve a constrained linear sum assignment problem for each entry.

    The output of this function is equivalent to, but significantly more
    efficient than,

    >>> def costs(cost_matrix):
    ...     total_costs = np.empty_like(cost_matrix)
    ...     num_rows, num_cols = cost_matrix.shape
    ...     for i in range(num_rows):
    ...         for j in range(num_cols):
    ...             total_costs[i, j] = clap.cost(i, j, cost_matrix)
    ...     return total_costs

    Parameters
    ----------
    cost_matrix : 2darray
        A matrix of costs.

    Returns
    -------
    2darray
        A matrix of total constrained lsap costs. The i, j entry of the matrix
        corresponds to the total lsap cost under the constraint that row i is
        assigned to column j.
    """
    cost_matrix = np.array(cost_matrix)
    cost_matrix = cost_matrix.astype(np.double)

    n_rows, n_cols = cost_matrix.shape
    if n_rows > n_cols:
        return costs(cost_matrix.T).T

    # Find the best lsap assignment from rows to columns without constrains.
    # Since there are at least as many columns as rows, row_idxs should
    # be identical to np.arange(n_rows). We depend on this.
    row_idxs = np.arange(n_rows)
    try:
        row4col, col4row, u, v = _solve(cost_matrix)
    except ValueError as e:
        if str(e) == "cost matrix is infeasible":
            return np.full((n_rows, n_cols), np.inf)
        else:
            raise e

    # Column vector of costs of each assignment in the lsap solution.
    lsap_costs = cost_matrix[row_idxs, col4row]
    lsap_total_cost = lsap_costs.sum()

    # Find the two minimum-cost columns for each row
    best_col_idxs = np.argmin(cost_matrix, axis=1)
    _cost_matrix = cost_matrix.copy()
    _cost_matrix[row_idxs, best_col_idxs] = np.inf
    second_best_col_idxs = np.argmin(_cost_matrix, axis=1)
    _cost_matrix[row_idxs, second_best_col_idxs] = np.inf
    third_best_col_idxs = np.argmin(_cost_matrix, axis=1)

    # When a row has its column stolen by a constraint, these are the columns
    # that might come into play when we are forced to resolve the assignment.
    if n_rows < n_cols:
        unused = np.setdiff1d(np.arange(n_cols), col4row)
        first_unused = np.argmin(cost_matrix[:, unused], axis=1)
        potential_cols = np.union1d(col4row, unused[first_unused])
    else:
        potential_cols = np.arange(n_cols)

    # When we add the constraint assigning row i to column j, lsap_col_idxs[i]
    # is freed up. If lsap_col_idxs[i] cannot improve on the cost of one of the
    # other row assignments, it does not need to be reassigned to another row.
    # If additionally column j is not in lsap_col_idxs, it is not taken away
    # from any of the other row assignments. In this situation, the resulting
    # total assignment costs are:
    total_costs = lsap_total_cost - lsap_costs[:, None] + cost_matrix

    for i, freed_j in enumerate(col4row):
        # When row i is constrained to another column, can column j be
        # reassigned to improve the assignment cost of one of the other rows?
        # To deal with that, we solve the lsap with row i omitted. For the
        # majority of constraints on row i's assignment, this will not conflict
        # with the constraint. When it does conflict, we fix the issue later.
        sub_ind = ~one_hot(i, n_rows)
        sub_cost_matrix = cost_matrix[sub_ind, :]

        new_row4col, new_col4row, new_u, new_v = lap.solve_lsap_with_removed_row(
            cost_matrix, i, row4col, col4row, u, v, modify_val=False
        )

        sub_total_cost = cost_matrix[sub_ind, new_col4row[sub_ind]].sum()
        new_col4row[i] = -1  # Row i is having a constraint applied.

        # This calculation will end up being wrong for the columns in
        # lsap_col_idxs[sub_col_ind]. This is because the constraint in
        # row i in these columns will conflict with the sub assignment.
        # These miscalculations are corrected later.
        total_costs[i, :] = cost_matrix[i, :] + sub_total_cost

        # new_col4row now contains the optimal assignment columns ignoring row i.
        new_col4row[i] = np.setdiff1d(col4row, new_col4row)[0]
        total_costs[i, new_col4row[i]] = cost_matrix[row_idxs, new_col4row].sum()

        # When we solve the lsap with row i removed, we update row4col accordingly.
        sub_row4col = new_row4col.copy()
        sub_row4col[sub_row4col == i] = -1
        sub_row4col[sub_row4col > i] -= 1

        for other_i, stolen_j in enumerate(new_col4row):
            if other_i == i:
                continue

            if not np.isfinite(cost_matrix[i, stolen_j]):
                total_costs[i, stolen_j] = cost_matrix[i, stolen_j]
                continue

            # Row i steals column stolen_j from other_i because of constraint.
            new_col4row[i] = stolen_j

            # Row other_i must find a new column. What is its next best option?
            best_j, second_best_j, third_best_j = (
                best_col_idxs[other_i],
                second_best_col_idxs[other_i],
                third_best_col_idxs[other_i],
            )

            # Note: Problem might occur if we have two j's that are both next
            # best, but one is not in col_idxs and the other is in col_idxs.
            # In this case, choosing the one not in col_idxs does not necessarily
            # give us the optimal assignment.
            # TODO: make the following if-else prettier.

            if (
                best_j != stolen_j
                and best_j not in new_col4row
                and (
                    cost_matrix[other_i, best_j] != cost_matrix[other_i, second_best_j]
                    or second_best_j not in new_col4row
                )
            ):
                new_col4row[other_i] = best_j
                total_costs[i, stolen_j] = cost_matrix[row_idxs, new_col4row].sum()
            elif second_best_j not in new_col4row and (
                cost_matrix[other_i, second_best_j]
                != cost_matrix[other_i, third_best_j]
                or third_best_j not in new_col4row
            ):
                new_col4row[other_i] = second_best_j
                total_costs[i, stolen_j] = cost_matrix[row_idxs, new_col4row].sum()
            else:
                # Otherwise, solve the lsap with stolen_j removed
                sub_sub_cost_matrix = sub_cost_matrix[:, potential_cols]
                sub_j = np.argwhere(potential_cols == stolen_j)[0][0]
                sub_new_col4row = new_col4row[sub_ind]
                sub_new_col4row = np.where(
                    sub_new_col4row.reshape(sub_new_col4row.size, 1) == potential_cols
                )[1]

                _, new_new_col4row, _, _ = lap.solve_lsap_with_removed_col(
                    sub_sub_cost_matrix,
                    sub_j,
                    sub_row4col[potential_cols],
                    sub_new_col4row,
                    new_u[sub_ind],  # dual variable associated with rows
                    new_v[potential_cols],  # dual variable associated with cols
                    modify_val=False,
                )

                total_costs[i, stolen_j] = (
                    cost_matrix[i, stolen_j]
                    + sub_sub_cost_matrix[np.arange(n_rows - 1), new_new_col4row].sum()
                )

            # Give other_i its column back in preparation for the next round.
            new_col4row[other_i] = stolen_j
            new_col4row[i] = -1

    # For those constraints which are compatible with the unconstrained lsap:
    total_costs[row_idxs, col4row] = lsap_total_cost

    return total_costs



import argparse

import numpy as np
import pyperf


def get_solvers():
    from laptools.clap import costs as costs_old
    from laptools.clap_new import costs as costs_new

    return {
        "clap_old": costs_old,
        "clap_new": costs_new,
    }


def time_func(n_inner_loops, solver, shape):
    cost_matrix = np.random.random(shape)
    t0 = pyperf.perf_counter()
    for i in range(n_inner_loops):
        solver(cost_matrix)
    return pyperf.perf_counter() - t0


def get_bench_name(size, solver_name):
    return "{}x{}-{}".format(size[0], size[1], solver_name)


def parse_args(benchopts):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--min-row-size-pow",
        type=int,
        metavar="POW",
        default=1,
        help="Smallest number of rows is 2^POW.",
    )
    parser.add_argument(
        "--min-col-size-pow",
        type=int,
        metavar="POW",
        default=1,
        help="Smallest number of cols is 2^POW.",
    )
    parser.add_argument(
        "--max-row-size-pow",
        type=int,
        metavar="POW",
        default=2,
        help="Largest number of rows is 2^POW.",
    )
    parser.add_argument(
        "--max-col-size-pow",
        type=int,
        metavar="POW",
        default=2,
        help="Largest number of cols is 2^POW.",
    )
    return parser.parse_args(benchopts)


def add_cmdline_args(cmd, args):
    cmd.append("--")
    cmd.extend(args.benchopts)


def main():
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.argparser.add_argument("benchopts", nargs="*")
    args = parse_args(runner.parse_args().benchopts)

    solvers = get_solvers()
    sizes = [
        (n_rows, n_cols)
        for n_rows in 2 ** np.arange(args.min_row_size_pow, args.max_row_size_pow + 1)
        for n_cols in 2 ** np.arange(args.min_col_size_pow, args.max_col_size_pow + 1)
    ]
    for size in sizes:
        for solver_name, solver_func in solvers.items():
            bench_name = get_bench_name(size, solver_name)
            runner.bench_time_func(bench_name, time_func, solver_func, size)


if __name__ == "__main__":
    main()



import numpy as np

from _augment import _solve, augment


def solve(cost_matrix, maximize=False):
    """Solve the linear sum assignment based on the cost matrix. The return
       value is the same as scipy.optimize.linear_sum_assignment.

    Parameters
    ----------
    cost_matrix : 2darray
         A matrix of costs.

    Returns
    -------
    row_ind, col_ind : array
        An array of row indices and one of corresponding column indices giving
        the optimal assignment. The cost of the assignment can be computed
        as ``cost_matrix[row_ind, col_ind].sum()``. The row indices will be
        sorted; in the case of a square cost matrix they will be equal to
        ``numpy.arange(cost_matrix.shape[0])``.
    """
    cost_matrix = np.array(cost_matrix)

    # The following are taken from scipy implementation.
    if len(cost_matrix.shape) != 2:
        raise ValueError(
            "expected a matrix (2-d array), got a %r array" % (cost_matrix.shape,)
        )

    if not (
        np.issubdtype(cost_matrix.dtype, np.number)
        or cost_matrix.dtype == np.dtype(np.bool)
    ):
        raise ValueError(
            "expected a matrix containing numerical entries, got %s"
            % (cost_matrix.dtype,)
        )

    if maximize:
        cost_matrix = -cost_matrix

    if np.any(np.isneginf(cost_matrix) | np.isnan(cost_matrix)):
        raise ValueError("matrix contains invalid numeric entries")

    cost_matrix = cost_matrix.astype(np.double)
    a = np.arange(np.min(cost_matrix.shape))

    # If the cost_matrix has more rows than columns
    if cost_matrix.shape[1] < cost_matrix.shape[0]:
        # Here, col4row holds the rows in cost_matrix that are in the assignment
        _, col4row, _, _ = _solve(cost_matrix.T)

        # Sort the row indexes in the assignment
        idx_sorted = np.argsort(col4row)

        return (col4row[idx_sorted], a[idx_sorted])
    # If the cost_matrix has more columns than rows
    else:
        _, col4row, _, _ = _solve(cost_matrix)
        return (a, col4row)


def solve_lsap_with_removed_row(
    cost_matrix, row_removed, row4col, col4row, u, v, modify_val=True
):
    """Solve the sub linear sum assignment problem with one row removed.

    Note: the removed row will still be assigned to a column.
    Note: While the cost matrix will not be modified, the dual variables would
          be updated as if the costs of removed row are uniformly zero.

    Parameters
    ----------
    cost_matrix : 2darray
         A matrix of costs.
    row_removed : int
         The index of the row that is to be removed.
    row4col: 1darray
             Specify the rows to which each column is matched with.
    col4row: 1darray
             Specify the columns to which each row is matched with.
    u : 1darray
        The dual cost vector for rows.
    v : 1darray
        The dual cost vector for columns.
    modify_val : bool, optional
        A flag that indicates whether variables are modified in place.
    """
    # cost_matrix = np.array(cost_matrix)
    n_rows, n_cols = cost_matrix.shape

    # Copy the variables if they are not modified in place.
    if not modify_val:
        row4col, col4row, u, v = row4col.copy(), col4row.copy(), u.copy(), v.copy()

    # If the freed up column does not contribute to lowering the costs of any
    # other rows, simply return the current assignments.
    freed_col = col4row[row_removed]
    freed_col_costs = cost_matrix[:, freed_col]
    if np.all(freed_col_costs >= cost_matrix[np.arange(n_rows), col4row]):
        return row4col, col4row, u, v

    # Update the cost matrix to reflect the row removal.
    # Create a copy of the cost matrix and update all costs associated with
    # row_removed to be zero. We don't modify the original cost values.
    cost_matrix_copy = cost_matrix.copy()
    cost_matrix_copy[row_removed] = 0

    # Update the dual variables
    u[row_removed] = np.min(cost_matrix_copy[row_removed] - v)

    # Perform another augmenting step, only on the sub-cost-matrix that
    # involves the rows and columns in the original optimal assignment.
    # The row to be removed have all of its associated costs set to 0.
    sub_cost_matrix = cost_matrix_copy[:, col4row]
    sub_v = v[col4row]

    # Remove the assignment associated with the removed row. Note that in the
    # submatrix, the removed row and the freed up column have the same index.
    sub_col4row = np.arange(n_rows)
    sub_col4row[row_removed] = -1
    sub_row4col = np.arange(n_rows)
    sub_row4col[row_removed] = -1

    # Find the shortest augmenting path for the sub square lsap and augment.
    augment(sub_cost_matrix, row_removed, sub_row4col, sub_col4row, u, sub_v)

    # Update the original assignment
    # Note: update every variable that depends on col4row for indexing first.
    row4col[col4row] = sub_row4col
    v[col4row] = sub_v
    col4row[:] = col4row[sub_col4row]

    return row4col, col4row, u, v


def solve_lsap_with_removed_col(
    cost_matrix, col_removed, row4col, col4row, u, v, modify_val=True
):
    """Solve the sub linear sum assignment problem with one column removed.

    Note: While the cost matrix will not be modified, the dual variables would
          be updated as if the removed column does not exist.

    Parameters
    ----------
    cost_matrix : 2darray
         A matrix of costs.
    col_removed : int
         The index of the column that is to be removed.
    row4col: 1darray
             Specify the rows to which each column is matched with.
    col4row: 1darray
             Specify the columns to which each row is matched with.
    u : 1darray
        The dual cost vector for rows.
    v : 1darray
        The dual cost vector for columns.
    """
    row_freed = row4col[col_removed]

    # If the removed column is resigned to nothing
    # No need to reassign anything
    if row_freed == -1:
        return row4col, col4row, u, v

    # Copy the variables if they are not modified in place.
    if not modify_val:
        row4col, col4row, u, v = row4col.copy(), col4row.copy(), u.copy(), v.copy()

    # Update the cost matrix to reflect the column removal.
    # Create a copy of the cost matrix and update all costs associated with
    # col_removed to be infinity. We don't modify the original cost values.
    # Note: need to make sure that cost_matrix has dtype float.
    cost_matrix_copy = cost_matrix.copy()
    cost_matrix_copy[:, col_removed] = np.inf

    # Removed the assignment associated with the removed column.
    col4row[row_freed] = -1
    row4col[col_removed] = -1

    # Perform another augmenting step
    augment(cost_matrix_copy, row_freed, row4col, col4row, u, v)

    return row4col, col4row, u, v



"""
   isort:skip_file
"""

import random

import numpy as np
from scipy.optimize import linear_sum_assignment

from laptools._util import one_hot
from laptools import lap


def test_solve_lsap_with_removed_row():
    """Tests for solving linear sum assignments with one row removed."""
    num_rows = 10
    num_cols = 500
    num_rounds = 100

    for i in range(num_rounds):
        # Note that here we set all costs to integer values, which might
        # lead to existence of multiple solutions.
        cost_matrix = np.random.randint(10, size=(num_rows, num_cols))
        cost_matrix = cost_matrix.astype(np.double)

        removed_row = random.randint(0, num_rows - 1)
        row_idx_1, col_idx_1 = linear_sum_assignment(cost_matrix)

        # Get the submatrix with the removed row
        sub_cost_matrix = cost_matrix[~one_hot(removed_row, num_rows), :]
        sub_row_idx_1, sub_col_idx_1 = linear_sum_assignment(sub_cost_matrix)

        # Solve the problem with dynamic algorithm
        row4col, col4row, u, v = lap._solve(cost_matrix)
        assert (
            np.array_equal(col_idx_1, col4row)
            or cost_matrix[row_idx_1, col_idx_1].sum()
            == cost_matrix[row_idx_1, col4row].sum()
        )

        lap.solve_lsap_with_removed_row(
            cost_matrix, removed_row, row4col, col4row, u, v
        )
        assert (
            np.array_equal(sub_col_idx_1, col4row[~one_hot(removed_row, num_rows)])
            or sub_cost_matrix[sub_row_idx_1, sub_col_idx_1].sum()
            == cost_matrix[
                ~one_hot(removed_row, num_rows),
                col4row[~one_hot(removed_row, num_rows)],
            ].sum()
        )


def test_solve_lsap_with_removed_col():
    """Tests for solving linear sum assignments with one column removed."""
    num_rows = 10
    num_cols = 20
    num_rounds = 1000

    for i in range(num_rounds):
        cost_matrix = np.random.randint(10, size=(num_rows, num_cols))
        cost_matrix = cost_matrix.astype(np.double)

        row_idx_1, col_idx_1 = linear_sum_assignment(cost_matrix)
        # Note that here we specifically pick a column that appears in the
        # previous optimal assignment.
        removed_col = random.choice(col_idx_1)

        # Get the submatrix with the removed col
        sub_cost_matrix = cost_matrix[:, ~one_hot(removed_col, num_cols)]
        sub_row_idx_1, sub_col_idx_1 = linear_sum_assignment(sub_cost_matrix)
        sub_cost_matrix_sum = sub_cost_matrix[sub_row_idx_1, sub_col_idx_1].sum()
        for i in range(len(sub_col_idx_1)):
            if sub_col_idx_1[i] >= removed_col:
                # Need to increment 1 to return these to their original index
                sub_col_idx_1[i] += 1

        # Solve the problem with dynamic algorithm
        row4col, col4row, u, v = lap._solve(cost_matrix)
        assert (
            np.array_equal(col_idx_1, col4row)
            or cost_matrix[row_idx_1, col_idx_1].sum()
            == cost_matrix[row_idx_1, col4row].sum()
        )

        lap.solve_lsap_with_removed_col(
            cost_matrix, removed_col, row4col, col4row, u, v
        )
        assert (
            np.array_equal(sub_col_idx_1, col4row)
            or sub_cost_matrix_sum == cost_matrix[row_idx_1, col4row].sum()
        )



import numpy as np

from _augment import augment


def test_augment():
    """Augment should modify its arguments inplace."""
    cost_matrix = np.array([[0.0, 0.0, 0.0], [0.0, 2.0, 5.0], [2.0, 3.0, 2.0]])
    row4col = np.array([-1, 1, 2])
    col4row = np.array([-1, 1, 2])
    u = np.array([0.0, 2.0, 2.0])
    v = np.array([-2.0, 0.0, 0.0])

    row4col_copy, col4row_copy = row4col.copy(), col4row.copy()

    augment(cost_matrix, 0, row4col, col4row, u, v)

    assert row4col_copy.tolist() != row4col.tolist()
    assert col4row_copy.tolist() != col4row.tolist()



import random

import numpy as np


def uniform_matrix(shape, low=0.0, high=1.0):
    """Generate a uniformly random matrix of the given shape."""
    return np.random.uniform(low=low, high=high, size=shape)


def geometric_matrix(shape, low=0.0, high=1.0):
    """Generate a geometric matrix of the given shape."""
    n_rows, n_cols = shape
    A = np.random.uniform(low=low, high=high, size=(n_rows + n_cols,))
    B = np.random.uniform(low=low, high=high, size=(n_rows + n_cols,))

    A_mat = np.array(
        [[A[i] - A[n_rows + j] for j in range(n_cols)] for i in range(n_rows)]
    )
    B_mat = np.array(
        [[B[i] - B[n_rows + j] for j in range(n_cols)] for i in range(n_rows)]
    )

    return np.sqrt(A_mat ** 2 + B_mat ** 2) + 1.0


def machol_wien_matrix(shape):
    """Generate a Machol-Wien matrix of the given shape."""
    n_rows, n_cols = shape
    return np.array(
        [[i * j + 1 for j in range(1, n_cols + 1)] for i in range(1, n_rows + 1)]
    )


def random_machol_wien_matrix(shape, low=0.0, high=1.0):
    """Generate a random Machol-Wien matrix of the given shape."""
    n_rows, n_cols = shape
    return np.array(
        [
            [random.randint(1, i * j + 1) for j in range(1, n_cols + 1)]
            for i in range(1, n_rows + 1)
        ]
    )



"""Functions for solving variants of the linear assignment problem."""
import numpy as np


def one_hot(idx, length):
    """Return a 1darray of False with a single True in the idx'th entry.

    Example
    -------
    >>> one_hot(0, 3)
    array([ True, False, False])

    Parameters
    ----------
    idx : int
        Index of the desired nonzero element.
    length : int
        Length of the resulting array.

    Returns
    -------
    1darray(bool)
        A one-hot array of False values with a True in the idx'th entry.
    """
    one_hot = np.zeros(length, dtype=np.bool)
    one_hot[idx] = True
    return one_hot



import argparse

import matplotlib.pyplot as plt
import numpy as np
import pyperf


def get_solver_from_bench(bench):
    """Extract the solver from a benchmark.
    Assumes each benchmark's name is of the format:
        "n_rows n_cols solver_name"
    """
    _, solver = bench.get_name().split("-")
    return solver


def get_size_from_bench(bench):
    """Extract the problem size from a benchmark.
    Assumes each benchmark's name is of the format:
        "n_rows n_cols solver_name"
    """
    size, _ = bench.get_name().split("-")
    n_rows, n_cols = size.split("x")
    return (int(n_rows), int(n_cols))


def get_solver_to_benches(suite):
    """Get a map from each solver to a list of its benchmarks."""
    solver_to_benches = {}
    for bench in suite:
        solver = get_solver_from_bench(bench)
        if solver not in solver_to_benches:
            solver_to_benches[solver] = []
        solver_to_benches[solver].append(bench)
    return solver_to_benches


def get_data_from_benches(benches):
    """Extract the problem size and runtimes from each benchmark."""
    sizes = [get_size_from_bench(bench) for bench in benches]
    times = [bench.get_values() for bench in benches]
    return np.array(sizes), np.array(times)


# TODO: can't do a loglog plot
def plot_suite(suite):
    """Plot the performance of each solver."""
    solver_to_benches = get_solver_to_benches(suite)
    for solver, benches in solver_to_benches.items():
        sizes, times = get_data_from_benches(benches)
        medians = np.median(times, axis=1)
        sizes_str = ["{}x{}".format(n_rows, n_cols) for n_rows, n_cols in sizes]
        plt.plot(sizes_str, medians, label=solver)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("suitefile")
    parser.add_argument("outputfile")
    return parser.parse_args()


def main():
    args = parse_args()
    suite = pyperf.BenchmarkSuite.load(args.suitefile)
    plot_suite(suite)
    plt.legend()
    plt.xlabel("Problem size")
    plt.ylabel("Time to solve")
    plt.savefig(args.outputfile)


if __name__ == "__main__":
    main()



# Included LSAP tests from Scipy
# Author: Brian M. Clapper, G. Varoquaux, Lars Buitinck
# License: BSD

import numpy as np
from numpy.testing import assert_array_equal
from pytest import raises as assert_raises
from scipy.optimize import linear_sum_assignment
from scipy.sparse.sputils import matrix

# from laptools import lap
import laptools
from laptools import lap


# fmt: off
def test_linear_sum_assignment():
    for sign in [-1, 1]:
        for cost_matrix, expected_cost in [
            # Square
            ([[400, 150, 400],
              [400, 450, 600],
              [300, 225, 300]],
             [150, 400, 300]
             ),

            # Rectangular variant
            ([[400, 150, 400, 1],
              [400, 450, 600, 2],
              [300, 225, 300, 3]],
             [150, 2, 300]),

            # Square
            ([[10, 10, 8],
              [9, 8, 1],
              [9, 7, 4]],
             [10, 1, 7]),

            # Rectangular variant
            ([[10, 10, 8, 11],
              [9, 8, 1, 1],
              [9, 7, 4, 10]],
             [10, 1, 4]),

            # n == 2, m == 0 matrix
            ([[], []],
             []),

            # Square with positive infinities
            ([[10, float("inf"), float("inf")],
              [float("inf"), float("inf"), 1],
              [float("inf"), 7, float("inf")]],
             [10, 1, 7]),
        ]:

            maximize = (sign == -1)
            cost_matrix = sign * np.array(cost_matrix)
            expected_cost = sign * np.array(expected_cost)

            row_ind, col_ind = lap.solve(cost_matrix, maximize=maximize)
            assert_array_equal(row_ind, np.sort(row_ind))
            assert_array_equal(expected_cost, cost_matrix[row_ind, col_ind])

            cost_matrix = cost_matrix.T
            row_ind, col_ind = lap.solve(cost_matrix, maximize=maximize)
            assert_array_equal(row_ind, np.sort(row_ind))
            assert_array_equal(np.sort(expected_cost),
                               np.sort(cost_matrix[row_ind, col_ind]))


def test_linear_sum_assignment_input_validation():
    # Since the input should be a 2D array.
    assert_raises(ValueError, lap.solve, [1, 2, 3])

    cost_matrix = [[1, 2, 3], [4, 5, 6]]
    assert_array_equal(lap.solve(cost_matrix),
                       lap.solve(np.asarray(cost_matrix)))
    assert_array_equal(lap.solve(cost_matrix),
                       lap.solve(matrix(cost_matrix)))

    eye = np.identity(3)
    assert_array_equal(lap.solve(eye.astype(np.bool)),
                       lap.solve(eye))
    assert_raises(ValueError, lap.solve, eye.astype(str))

    eye[0][0] = np.nan
    assert_raises(ValueError, lap.solve, eye)

    eye = np.identity(3)
    eye[1][1] = -np.inf
    assert_raises(ValueError, lap.solve, eye)

    eye = np.identity(3)
    eye[:, 0] = np.inf
    assert_raises(ValueError, lap.solve, eye)

    # Note that solve requires that num_rows cannot exceed num_cols
    cost_matrix = np.array([[0     , np.inf, np.inf],
                            [np.inf,      0, np.inf],
                            [np.inf, np.inf,      0],
                            [0     ,      0,      0]])
    assert_raises(ValueError, laptools._solve, cost_matrix)

# fmt: on


def test_linear_sum_assignment_against_scipy():
    for i in range(100):
        cost_matrix = np.random.rand(5, 10)

        row_ind_1, col_ind_1 = lap.solve(cost_matrix)
        row_ind_2, col_ind_2 = linear_sum_assignment(cost_matrix)

        assert_array_equal(row_ind_1, row_ind_2)
        assert_array_equal(col_ind_1, col_ind_2)

        cost_matrix = cost_matrix.T

        row_ind_1, col_ind_1 = lap.solve(cost_matrix)
        row_ind_2, col_ind_2 = linear_sum_assignment(cost_matrix)

        assert_array_equal(row_ind_1, row_ind_2)
        assert_array_equal(col_ind_1, col_ind_2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import sys

import setuptools
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

with open("README.md") as readme_file:
    readme = readme_file.read()

test_requirements = [
    "codecov",
    "pytest",
    "pytest-cov",
]

docs_requirements = [
    "sphinx==1.8.5",
]

setup_requirements = [
    "pytest-runner",
    "pybind11>=2.5.0",
]

perf_requirements = [
    "pyperf",
    "matplotlib",
    "numpy",
    "scipy",
    "munkres",
    "lap",
    "lapsolver",
    "lapjv",
]

dev_requirements = [
    *test_requirements,
    *docs_requirements,
    *setup_requirements,
    *perf_requirements,
    "pre-commit",
    "bumpversion>=0.5.3",
    "ipython>=7.5.0",
    "tox>=3.5.2",
    "twine>=1.13.0",
    "wheel>=0.33.1",
]

requirements = [
    "numpy",
    "scipy",
    "numba",
]

extra_requirements = {
    "test": test_requirements,
    "docs": docs_requirements,
    "setup": setup_requirements,
    "dev": dev_requirements,
    "perf": perf_requirements,
    "all": [
        *requirements,
        *test_requirements,
        *docs_requirements,
        *setup_requirements,
        *dev_requirements,
        *perf_requirements,
    ],
}


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __str__(self):
        import pybind11

        return pybind11.get_include()


# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    import os

    with tempfile.NamedTemporaryFile("w", suffix=".cpp", delete=False) as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        fname = f.name
    try:
        compiler.compile([fname], extra_postargs=[flagname])
    except setuptools.distutils.errors.CompileError:
        return False
    finally:
        try:
            os.remove(fname)
        except OSError:
            pass
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14/17] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """
    flags = ["-std=c++17", "-std=c++14", "-std=c++11"]

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError("Unsupported compiler -- at least C++11 support " "is needed!")


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {
        "msvc": ["/EHsc"],
        "unix": [],
    }
    l_opts = {
        "msvc": [],
        "unix": [],
    }

    if sys.platform == "darwin":
        darwin_opts = ["-stdlib=libc++", "-mmacosx-version-min=10.7"]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == "unix":
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")

        for ext in self.extensions:
            ext.define_macros = [
                ("VERSION_INFO", '"{}"'.format(self.distribution.get_version()))
            ]
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


setup(
    author="Jacob Moorman",
    author_email="jacob@moorman.me",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research ",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
    cmdclass={"build_ext": BuildExt},
    description="Fast constrained linear assignment problem (CLAP) solvers",
    ext_modules=[
        Extension(
            "_augment",
            sorted(
                ["src/cpp/_augment.cpp"]
            ),  # Sort input source files to ensure bit-for-bit reproducible builds
            include_dirs=[get_pybind_include()],  # Path to pybind11 headers
            language="c++",
        ),
    ],
    extras_require=extra_requirements,
    install_requires=requirements,
    license="MIT License",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="laptools",
    name="laptools",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.6",
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/jdmoorman/laptools",
    # Do not edit this string manually, always use bumpversion
    # Details in CONTRIBUTING.rst
    version="0.2.2",
    zip_safe=False,
)
